import py.test
from fudge import patch as fudge_patch, Fake
from fudge.inspector import arg
from pyramid import testing
from pyramid.httpexceptions import HTTPNotFound, HTTPServerError
from sqlalchemy.orm import exc
from httpretty import HTTPretty
from belt.utils import pypi_url
from belt import models
from belt.axle import split_package_name

import functools
import sys


def wraps(func):
    def inner(f):
        f = functools.wraps(func)(f)
        original = getattr(func, '__wrapped__', func)
        f.__wrapped__ = original
        return f
    return inner


class patch(fudge_patch):

    def __call__(self, fn):

        @wraps(fn)
        def caller(*args, **kw):
            fakes = self.__enter__()
            if not isinstance(fakes, (tuple, list)):
                fakes = [fakes]
            args += tuple(fakes)
            value = None
            try:
                value = fn(*args, **kw)
            except:
                etype, val, tb = sys.exc_info()
                self.__exit__(etype, val, tb)
                raise etype, val, tb
            else:
                self.__exit__(None, None, None)
            return value

        # py.test uses the length of mock.patchings to determine how many
        # arguments to ignore when performing its dependency injection
        if not hasattr(caller, 'patchings'):
            caller.patchings = []
        caller.patchings.extend([1 for path in self.obj_paths])
        return caller


pypi_base_url = 'https://pypi.python.org/packages'


def create_package(path, content=u''):
    import os

    if content:
        filename = str(path)
        path.write(content)
    else:
        filename = u''
    basename = os.path.basename(str(path))
    name, version = split_package_name(basename)

    package = models.Package(name=name)
    rel = models.Release(version=version)
    package.releases.append(rel)
    rel.files.append(models.File(filename=filename, md5='gg'))
    return package


class TestDownloadPackage(object):

    def test_returns_404_if_package_not_on_pypi(self, http, db_session, dummy_request):
        from ..views import download_package

        request = dummy_request
        url = pypi_url(pypi_base_url, 'source', 'flake8', 'flake8-2.0.tar.gz')
        HTTPretty.register_uri(HTTPretty.GET, url, body='Missed it!', status=404)
        request.matchdict = {'package': 'flake8', 'kind': 'source',
                             'letter': 'f', 'basename': 'flake8-2.0.tar.gz'}

        with py.test.raises(HTTPNotFound):
            download_package(request)

    def test_returns_500_when_server_error(self, http, dummy_request):
        from ..views import download_package

        request = dummy_request
        url = pypi_url(pypi_base_url, 'source', 'flake8', 'flake8-2.0.tar.gz')
        HTTPretty.register_uri(HTTPretty.GET, url, body='Yikes!', status=500)
        request.matchdict = {'package': 'flake8', 'kind': 'source',
                             'letter': 'f', 'basename': 'flake8-2.0.tar.gz'}

        with py.test.raises(HTTPServerError):
            download_package(request)

    def test_obtains_missing_file_from_pypi(self, tmpdir, http, db_session,
                                            dummy_request):
        from ..views import download_package

        pkg = tmpdir.join('foo-1.2.tar.gz')
        package = create_package(pkg, content=u'Short lived')

        # remove just the package file, not the record
        pkg.remove()
        db_session.add(package)
        db_session.flush()

        url = pypi_url(pypi_base_url, 'source', 'foo', 'foo-1.2.tar.gz')
        HTTPretty.register_uri(HTTPretty.GET, url, body='GOT IT', status=200)

        dummy_request.matchdict = {'package': 'foo', 'kind': 'source',
                                   'letter': 'f', 'basename': 'foo-1.2.tar.gz'}
        response = download_package(dummy_request)
        assert 'GOT IT' == response.body


class TestRequestNonExistentPackage(object):

    def test_creates_package_record(self, http, db_session, dummy_request):
        from ..views import download_package

        url = pypi_url(pypi_base_url, 'source', 'foo', 'foo-1.2.tar.gz')
        HTTPretty.register_uri(HTTPretty.GET, url, body='GOT IT', status=200)

        request = testing.DummyRequest()
        request.matchdict = {'package': 'foo', 'kind': 'source',
                             'letter': 'f', 'basename': 'foo-1.2.tar.gz'}
        download_package(request)

        db_session.query(models.Package).filter_by(name='foo').one()

    def test_creates_release_record(self, http, db_session, dummy_request):
        from ..views import download_package

        url = pypi_url(pypi_base_url, 'source', 'foo', 'foo-1.2.tar.gz')
        HTTPretty.register_uri(HTTPretty.GET, url, body='GOT IT', status=200)

        dummy_request.matchdict = {'package': 'foo', 'kind': 'source',
                                   'letter': 'f', 'basename': 'foo-1.2.tar.gz'}
        download_package(dummy_request)

        models.Release.for_package('foo', '1.2')

    def test_creates_file_record(self, http, db_session, dummy_request):
        from ..views import download_package

        url = pypi_url(pypi_base_url, 'source', 'foo', 'foo-1.2.tar.gz')
        HTTPretty.register_uri(HTTPretty.GET, url, body='GOT IT', status=200)

        dummy_request.matchdict = {'package': 'foo', 'kind': 'source',
                                   'letter': 'f', 'basename': 'foo-1.2.tar.gz'}
        download_package(dummy_request)

        models.File.for_release('foo', '1.2')


class TestRequestNonExistentFile(object):

    def test_redirects_to_pypi(self, http, db_session, dummy_request):
        from ..views import download_package

        url = pypi_url(pypi_base_url, 'source', 'foo', 'foo-1.2.tar.gz')
        HTTPretty.register_uri(HTTPretty.GET, url, body='GOT IT', status=200)

        dummy_request.matchdict = {'package': 'foo', 'kind': 'source',
                                   'letter': 'f', 'basename': 'foo-1.2.tar.gz'}
        response = download_package(dummy_request)
        assert 'GOT IT' == response.body


class TestPackageList(object):

    def test_requests_package_data_from_pypi(self, dummy_request):
        from ..views import package_list
        dummy_request.matchdict = {'package': u'foo'}

        pkg = Fake('Package').has_attr(name='foo', releases=[1])

        with patch('belt.views.Package') as Package:
            Package.expects('by_name').with_args('foo').raises(exc.NoResultFound)
            (Package.expects('create_from_pypi')
             .with_args(name='foo', package_dir=arg.any())
             .returns(pkg))
            package_list(dummy_request)

    @patch('belt.views.DBSession')
    @patch('belt.views.Package')
    @patch('belt.views.package_releases')
    def test_requests_package_releases_if_none_exist(self, DBSession, Package,
                                                     package_releases,
                                                     dummy_request):
        from ..views import package_list

        rel = Fake('Release').has_attr(version='106')
        package_releases.expects_call().with_args('quux', location=arg.any()).returns([rel])

        pkg = (Fake('pkg').has_attr(name='quux', releases=[]))
        Package.expects('by_name').with_args('quux').returns(pkg)
        DBSession.expects('add').with_args(pkg)

        dummy_request.matchdict = {'package': u'quux'}
        package_list(dummy_request)



import os
import sys
import importlib
from pyramid.scripts import pshell


class Shell(pshell.PShellCommand):

    def make_shell(self):
        try:
            start = os.environ['PYTHONSTARTUP']
        except KeyError:
            return super(Shell, self).make_shell()

        dirname = os.path.dirname(start)
        module = os.path.basename(start)
        module, _ = os.path.splitext(module)
        sys.path.insert(0, dirname)
        mod = importlib.import_module(module, os.path.basename(dirname))
        sys.path.pop(0)

        def shell(env, help):
            code = mod.EditableBufferInteractiveConsole(locals=env)
            code.interact(help + "\n\n")
        return shell


def pshell():
    pass


if __name__ == "__main__":
    command = Shell(sys.argv)
    command.run()



import re
import os
import urllib2
import logging
import urlparse
from .values import Path, ReleaseValue


logger = logging.getLogger(__name__)


def get_url(url):
    try:
        fo = urllib2.urlopen(url)
    except urllib2.URLError as err:
        if hasattr(err, 'reason'):
            msg = u"Failed to connect to server because: " + err.reason
        if hasattr(err, 'code'):
            msg = u'Failed to complete response with error code: %d' % err.code
        logger.exception(msg)
        raise
    return fo


def local_packages(packages_root):
    for item in os.listdir(packages_root):
        if os.path.isdir(os.path.join(packages_root, item)):
            yield item


def pypi_versions(package_page, url=None):
    import lxml.html
    html = lxml.html.fromstring(package_page)
    if url:
        html.make_links_absolute(url)
    links = []
    for elem, attr, link, pos in html.iterlinks():
        if elem.tag != 'a':
            continue
        if elem.attrib.get('rel', '') == 'homepage':
            continue
        links.append(lxml.html.tostring(elem))
    return links


def pypi_package_page(url):
    url = convert_url_to_pypi(url)
    return get_url(url).read()


def convert_url_to_pypi(url):
    # TODO define pypi url once only
    base = 'https://pypi.python.org'
    return urlparse.urljoin(base, urlparse.urlsplit(url).path)


def local_releases(packages_root, package_name):
    package_dir = os.path.join(packages_root, package_name)
    candidates = os.listdir(package_dir) if os.path.exists(package_dir) else []
    return [ReleaseValue(version, package_dir) for version in candidates if not version.endswith('.md5')]


def get_package(packages_root, package_name, package_version):
    return Path(os.path.join(packages_root, package_name, package_version))


def get_package_from_pypi(url):
    return get_url(url)


def pypi_url(pypi_base_url, kind, package_name, basename):
    return '{}/{}/{}/{}/{}'.format(pypi_base_url.rstrip('/'), kind,
                                   package_name[0],
                                   package_name, basename)


def store_locally(path, fo):
    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    with open(path, 'w') as package:
        package.write(fo.read())


def get_search_names(name):
    parts = re.split('[-_]', name)
    if len(parts) == 1:
        return parts

    result = set()
    for i in range(len(parts) - 1, 0, -1):
        for s1 in '-_':
            prefix = s1.join(parts[:i])
            for s2 in '-_':
                suffix = s2.join(parts[i:])
                for s3 in '-_':
                    result.add(s3.join([prefix, suffix]))
    return list(result)



from pyramid.config import Configurator
from pyramid_jinja2 import renderer_factory
from sqlalchemy import engine_from_config
from .models import DBSession, Base


def main(global_config, **settings):
    """ This function returns a WSGI application.

    It is usually called by the PasteDeploy framework during
    ``paster serve``.
    """
    engine = engine_from_config(settings, 'sqlalchemy.')
    DBSession.configure(bind=engine)
    Base.metadata.bind = engine

    settings = dict(settings)
    settings.setdefault('jinja2.i18n.domain', 'belt')

    config = Configurator(settings=settings)
    config.add_translation_dirs('locale/')
    config.include('pyramid_jinja2')
    config.include('pyramid_tm')

    config.add_static_view('static', 'static')
    config.add_route('simple', '/simple/')
    config.add_route('package_list', '/simple/{package}/')
    config.add_route('package_version', '/simple/{package}/{version}')
    config.add_route('download_package',
                     '/packages/{kind}/{letter}/{package}/{basename}')

    config.scan('belt', ignore='belt.tests')
    config.add_renderer('.html', renderer_factory)

    return config.make_wsgi_app()



import os
import pytest
from belt import models
from httpretty import HTTPretty
from pyramid.paster import get_appsettings
from pyramid import testing
from sqlalchemy import engine_from_config


ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..'))


@pytest.fixture
def http(request):
    HTTPretty.reset()
    HTTPretty.enable()
    request.addfinalizer(HTTPretty.disable)


@pytest.fixture(scope='session')
def connection(request):
    settings = get_appsettings(os.path.join(ROOT, 'test.ini'))
    engine = engine_from_config(settings, 'sqlalchemy.')
    models.Base.metadata.create_all(engine)
    connection = engine.connect()
    models.DBSession.registry.clear()
    models.DBSession.configure(bind=connection)
    models.Base.metadata.bind = engine
    request.addfinalizer(models.Base.metadata.drop_all)
    return connection


@pytest.fixture
def db_session(request, connection):
    from transaction import abort
    trans = connection.begin()
    request.addfinalizer(trans.rollback)
    request.addfinalizer(abort)

    from belt.models import DBSession
    return DBSession


@pytest.fixture
def config(request):
    from pyramid import testing  # noqa
    config = testing.setUp()
    request.addfinalizer(testing.tearDown)
    return config


@pytest.fixture
def dummy_request(config, tmpdir):
    config.add_settings({'local_packages': str(tmpdir)})
    config.manager.get()['request'] = request = testing.DummyRequest()
    return request



import os
import re
import glob
import shutil
import urlparse
from belt.axle import split_package_name, mkdir_p


def _pipcache():
    import argparse
    desc = 'Copy PIP_DOWNLOAD_CACHE to pypi dir'
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('--pip-download-cache',
                        help='Directory packages can be found in',
                        required=True)
    parser.add_argument('--pypi-dir',
                        help='Directory to copy packages to',
                        required=True)
    args = parser.parse_args()
    pipcache(args.pip_download_cache, args.pypi_dir)


def pipcache(pip_download_cache, pypi_dir):
    pip_download_cache = pip_download_cache.rstrip('/')
    md5_re = re.compile(r'\?md5=\w*$')
    for fn in glob.iglob(pip_download_cache + '/*'):
        real_name = os.path.basename(urlparse.unquote(fn))
        if not real_name or real_name.endswith('.content-type'):
            continue

        package_name, _ = split_package_name(real_name)
        dest = os.path.join(pypi_dir, package_name, real_name)

        dest = md5_re.sub('', dest)
        mkdir_p(os.path.dirname(dest))
        shutil.copyfile(fn, dest)


if __name__ == "__main__":
    _pipcache()



import os
import logging
from .axle import split_package_name


logger = logging.getLogger(__name__)


class Path(object):

    def __init__(self, path):
        self.path = path

    @property
    def exists(self):
        return os.path.exists(self.path)


class ReleaseValue(object):

    _md5 = ''

    def __init__(self, name, package_dir):
        self.name, self.number = split_package_name(name)
        self.fullname = name

        self.package_dir = package_dir

    def __eq__(self, other):
        return self.fullname == other

    def __repr__(self):
        return self.fullname

    @property
    def md5(self):
        if not self._md5:

            hash_name = self.fullpath + '.md5'
            try:
                with open(hash_name) as hashed:
                    self._md5 = hashed.read()
            except IOError:
                # msg = u'{} does not exist'.format(hash_name)
                # logger.exception(msg)
                pass
        return self._md5

    @property
    def fullpath(self):
        return os.path.join(self.package_dir, self.fullname)



import os
import logging
from hashlib import md5
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import (scoped_session, sessionmaker, relationship,
                            joinedload)
from zope.sqlalchemy import ZopeTransactionExtension

from sqlalchemy.ext.hybrid import Comparator, hybrid_property
from sqlalchemy import (Column, Integer, Text, Boolean, ForeignKey, DateTime,
                        func, UniqueConstraint, or_)
from .utils import local_packages, local_releases, get_search_names
from .axle import get_package_name, split_package_name


DBSession = scoped_session(sessionmaker(extension=ZopeTransactionExtension()))
Base = declarative_base()


log = logging.getLogger(__name__)


class CaseInsensitiveComparator(Comparator):
    def __eq__(self, other):
        return func.lower(self.__clause_element__()) == func.lower(other)


class Package(Base):

    __tablename__ = 'package'
    id = Column(Integer, primary_key=True)
    name = Column(Text, unique=True)
    releases = relationship('Release', backref='package',
                            cascade='all, delete-orphan', passive_deletes=True)

    def __init__(self, name, package_dir=None):
        self.name = name
        self.package_dir = package_dir

    def __repr__(self):
        return u'package: {}'.format(self.name)

    @hybrid_property
    def name_insensitive(self):
        return self.name.lower()

    @name_insensitive.comparator
    def name_insensitive(cls):
        return CaseInsensitiveComparator(cls.name)

    @classmethod
    def by_name(cls, name):
        names = get_search_names(name)
        or_query = or_(*[Package.name_insensitive == n for n in names])
        joined = joinedload(Package.releases, Release.files)
        return DBSession.query(Package).options(joined).filter(or_query).one()

    @classmethod
    def create_from_pypi(cls, name, package_dir):
        name = get_package_name(name)
        return cls(name=name, package_dir=package_dir)


class Release(Base):
    __tablename__ = 'release'
    __table_args__ = (UniqueConstraint('version', 'package_id'), )

    id = Column(Integer, primary_key=True)
    local = Column(Boolean, default=False)
    package_id = Column(Integer, ForeignKey('package.id', ondelete='CASCADE'))
    version = Column(Text)
    created = Column(DateTime(timezone=True), nullable=False,
                     server_default=func.now())
    modified = Column(DateTime(timezone=True), nullable=False,
                      server_default=func.now(), onupdate=func.now())
    files = relationship('File', backref='release')


    def __repr__(self):
        if self.package:
            pkg_name = self.package.name
        else:
            pkg_name = u'UNKNOWN'
        return u'release: {}-{}'.format(pkg_name, self.version)

    @classmethod
    def for_package(cls, package, version):
        return (DBSession.query(Release).join(Release.package)
                .filter(Release.version == version, Package.name == package)
                .one())


class File(Base):
    __tablename__ = 'file'
    __table_args__ = (UniqueConstraint('kind', 'release_id', 'filename'), )

    id = Column(Integer, primary_key=True)
    created = Column(DateTime(timezone=True), nullable=False,
                     server_default=func.now())
    modified = Column(DateTime(timezone=True), nullable=False,
                      server_default=func.now(), onupdate=func.now())
    release_id = Column(Integer, ForeignKey('release.id', ondelete='CASCADE'))
    location = Column(Text, nullable=False, default=u'')
    filename = Column(Text, nullable=False, default=u'')
    md5 = Column(Text, nullable=False, default=u'')
    kind = Column(Text, nullable=False, default='source')
    download_url = Column(Text, unique=True)

    @property
    def fullpath(self):
        return os.path.join(self.location, self.filename)

    @classmethod
    def for_release(cls, package, version):
        return (DBSession
                .query(File)
                .join(File.release, Release.package)
                .filter(Release.version == version, Package.name == package)
                .all())

    @classmethod
    def by_filename(cls, filename):
        name, version = split_package_name(filename)
        return (DBSession
                .query(File)
                .join(File.release, Release.package)
                .filter(Release.version == version,
                        Package.name_insensitive == name,
                        File.filename == filename)
                .one())


from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound


def get_or_create(session, model, defaults=None, **kwargs):
    query = session.query(model).filter_by(**kwargs)
    created = False
    try:
        instance = query.one()
    except NoResultFound:
        session.begin(nested=True)
        defaults = defaults or {}
        kwargs.update(defaults)
        instance = model(**kwargs)
        session.add(instance)
        try:
            session.flush()
        except IntegrityError:
            session.rollback()
            instance = query.one()
        else:
            created = True
    return instance, created


def seed_packages(package_dir):
    packages = []
    for pkg in local_packages(package_dir):
        releases = {}
        package = Package(name=pkg)
        for rel in local_releases(package_dir, pkg):
            root, ext = os.path.splitext(rel.fullpath)
            if ext in ['.md5']:
                continue

            release = releases.setdefault(rel.number,
                                          Release(version=rel.number))
            with open(rel.fullpath) as fo:
                hashed_content = md5(fo .read()).hexdigest()
            release.files.append(File(filename=rel.fullname,
                                      location=rel.package_dir, md5=hashed_content))
        for k, v in releases.items():
            package.releases.append(v)
        packages.append(package)
    return packages



import os.path
import urllib2
import cStringIO
import py.test
import lxml.html
from httpretty import HTTPretty, httprettified


def test_get_local_packages(tmpdir):
    from ..utils import local_packages

    tmpdir.mkdir('py.test')
    tmpdir.mkdir('yolk')
    tmpdir.mkdir('coverage')

    package_names = local_packages(str(tmpdir))
    assert ['coverage', 'py.test', 'yolk'] == sorted(list(package_names))


def test_returns_empty_list_if_dir_is_empty(tmpdir):
    from ..utils import local_packages
    package_names = local_packages(str(tmpdir))
    assert [] == list(package_names)


def test_get_dir_for_package(tmpdir):
    from ..utils import local_releases
    yolk = tmpdir.mkdir('yolk').join('yolk.1.0.1.tar.gz')
    yolk.write('1')
    package_dir = str(tmpdir)
    assert ['yolk.1.0.1.tar.gz'] == local_releases(package_dir, 'yolk')


def test_local_releases_excludes_hash_files(tmpdir):
    from ..utils import local_releases
    yolk = tmpdir.mkdir('yolk').join('yolk.1.0.1.tar.gz.md5')
    yolk.write('')
    package_dir = str(tmpdir)
    assert [] == local_releases(package_dir, 'yolk')


def test_local_releases_has_md5_attr(tmpdir):
    from ..utils import local_releases
    p = tmpdir.join('somename.exe')
    p.write('')

    p = tmpdir.join('somename.exe.md5')
    p.write('HASHED CONTENT')
    package_dir = str(tmpdir)
    local, = local_releases(package_dir, '')
    assert 'HASHED CONTENT' == local.md5


def test_returns_empty_list_if_no_local_releases(tmpdir):
    from ..utils import local_releases
    package_dir = str(tmpdir)
    assert [] == local_releases(package_dir, 'yolk')


def test_returns_empty_list_if_no_local_package_dir(tmpdir):
    from ..utils import local_releases
    package_dir = str(tmpdir)
    tmpdir.mkdir('yolk')
    assert [] == local_releases(package_dir, 'yolk')


def test_get_full_path_to_version(tmpdir):
    from ..utils import get_package
    yolk = tmpdir.mkdir('yolk').join('yolk.1.0.1.tar.gz')
    yolk.write('1')
    expected = os.path.join(str(tmpdir), 'yolk', 'yolk.1.0.1.tar.gz')
    package_dir = str(tmpdir)
    assert expected == get_package(package_dir, 'yolk',
                                   'yolk.1.0.1.tar.gz').path


class TestStoreLocally(object):

    def test_writes_file_content_to_destination(self, tmpdir):
        from ..utils import store_locally
        blub = tmpdir.mkdir('blub').join('blub.4.3.tar.gz')
        blub.write('1')

        fo = cStringIO.StringIO()
        fo.write('some code goes here')
        fo.seek(0)
        store_locally(str(blub), fo)

        with open(str(blub)) as output:
            assert 'some code goes here' == output.read()

    def test_makes_dir_if_nonexistant(self, tmpdir):
        from ..utils import store_locally
        destination = os.path.join(str(tmpdir), 'foo', 'bar', 'baz-1.zip')
        fo = cStringIO.StringIO()
        fo.write('path making')
        fo.seek(0)

        store_locally(destination, fo)

        with open(destination) as output:
            assert 'path making' == output.read()


class TestGetPackageFromPypi(object):

    @httprettified
    def test_get_package_from_pypi(self):
        from ..utils import get_package_from_pypi, pypi_url
        pypi_base_url = 'https://pypi.python.org/packages'
        url = pypi_url(pypi_base_url, 'source', 'flake8',
                       'flake8-2.0.tar.gz')

        HTTPretty.register_uri(HTTPretty.GET, url, body='Got it!')
        response = get_package_from_pypi(url)
        assert 'Got it!' == response.read()

    @httprettified
    def test_breaks_on_404(self):
        from ..utils import get_package_from_pypi, pypi_url
        pypi_base_url = 'https://pypi.python.org/packages'
        url = pypi_url(pypi_base_url, 'source', 'flake8',
                       'flake8-2.0.tar.gz')

        HTTPretty.register_uri(HTTPretty.GET, url, body='Missed it!',
                               status=404)

        with py.test.raises(urllib2.URLError):
            get_package_from_pypi(url)


def get_html_fixture(name):

    with open(os.path.join(os.path.dirname(__file__), name)) as fp:
        return fp.read()


class TestPypiVersions(object):

    def test_finds_all_links_in_page(self):
        from ..utils import pypi_versions
        page = '<p><a href="">a</a><a href="">b</a>'
        links = pypi_versions(page)
        assert 2 == len(links)

    def test_skips_homepage_links(self):
        from ..utils import pypi_versions
        page = '<p><a rel="homepage" href="">a</a><a href="">b</a>'
        links = ''.join(pypi_versions(page))
        html = lxml.html.fromstring(links)

        elems = []
        for elem, attr, link, post in html.iterlinks():
            if elem.attrib.get('rel', '') == 'homepage':
                elems.append(elem)
        assert 0 == len(elems)

    def test_makes_links_absolute(self):
        from ..utils import pypi_versions
        page = '<a href="../../packages/nose.tgz">nose</a>'
        link = pypi_versions(page, 'http://p.org/s/nose/')[0]
        assert link == '<a href="http://p.org/packages/nose.tgz">nose</a>'

    @httprettified
    def test_reads_pypi_version_of_package_page(self):
        from ..utils import pypi_package_page
        HTTPretty.register_uri(HTTPretty.GET,
                               'https://pypi.python.org/simple/nose/',
                               body='<a href="">G</a><a href="">B</a>',)

        response = pypi_package_page('http://localhost/simple/nose/')
        assert '<a href="">G</a><a href="">B</a>' == response



import os
import itertools
import fudge
import pytest
import py.test
from belt import models
from sqlalchemy import exc


def test_move_wheels_to_pypi_dir(tmpdir):
    from ..axle import copy_wheels_to_pypi

    wh = tmpdir.join('belt-0.2-py27-none-any.whl')
    wh.write('')

    pypi = tmpdir.mkdir('local')

    copy_wheels_to_pypi(wheel_dir=str(tmpdir), local_pypi=str(pypi))

    expected = os.path.join(str(pypi), 'belt', 'belt-0.2-py27-none-any.whl')
    assert os.path.exists(expected)


def test_doesnt_overwrite_existing_wheels(tmpdir):
    from ..axle import copy_wheels_to_pypi

    built_wheel = tmpdir.join('belt-0.2-py27-none-any.whl')
    built_wheel.write('')

    local_pypi = tmpdir.mkdir('local')
    package_wheel = local_pypi.mkdir('belt').join('belt-0.2-py27-none-any.whl')
    package_wheel.write('')

    with fudge.patch('belt.axle.shutil') as shutil:
        shutil.provides('copyfile').times_called(0)
        copy_wheels_to_pypi(wheel_dir=str(tmpdir), local_pypi=str(local_pypi))


@pytest.mark.parametrize(('package', 'name_and_version'), [
    ('bump-0.1.0.tar.gz', ('bump', '0.1.0')),
    ('fudge-22.1.4.zip', ('fudge', '22.1.4')),
    ('wheel-1.0.0a1.tar.bz2', ('wheel', '1.0.0a1')),
])
def test_split_package_name_detects_name(package, name_and_version):
    from ..axle import split_package_name
    assert name_and_version == split_package_name(package)


belt_pkg_data = [
    {'comment_text': '',
     'downloads': 133,
     'filename': 'belt-0.4.zip',
     'has_sig': False,
     'md5_digest': 'eea30ff3dc9b52c6658f7657c6dd10b5',
     'packagetype': 'sdist',
     'python_version': 'source',
     'size': 25988,
     'upload_time': '2002-01-01',
     'url': 'https://pypi.python.org/packages/source/b/belt/belt-0.4.zip'}]


class TestGetReleaseData(object):

    def test_sets_release_version(self):
        from ..axle import package_releases
        client = (fudge.Fake('client')
                  .expects('release_urls').returns(belt_pkg_data)
                  .expects('package_releases').returns(['0.1']))
        release, = package_releases('belt', client=client)
        assert u'0.1' == release.version

    def test_sets_release_file_md5(self):
        from ..axle import package_releases
        client = (fudge.Fake('client')
                  .expects('release_urls').returns(belt_pkg_data)
                  .expects('package_releases').returns(['0.1']))
        release, = package_releases('belt', client=client)
        assert 'eea30ff3dc9b52c6658f7657c6dd10b5' == release.files[0].md5

    def test_returns_list_of_releases_to_add_to_package(self, db_session):
        from ..axle import package_releases
        client = (fudge.Fake('client')
                  .expects('release_urls').returns(belt_pkg_data)
                  .expects('package_releases').returns(['0.1']))
        pkg = models.Package(name=u'belt')
        db_session.add(pkg)

        releases = package_releases('belt', client=client)
        pkg.releases.extend(list(releases))
        pkg = db_session.query(models.Package).filter_by(name='belt').one()
        assert 1 == len(pkg.releases)

    def test_release_list_does_not_guarantee_uniqueness(self, db_session):
        from ..axle import package_releases
        client = (fudge.Fake('client')
                  .expects('release_urls').returns(belt_pkg_data)
                  .expects('package_releases').returns(['0.1']))

        # add a package named belt with a 0.1 release
        pkg = models.Package(name=u'belt')
        pkg.releases.append(models.Release(version=u'0.1'))
        db_session.add(pkg)

        releases = package_releases('belt', client=client)

        # add the returned packages and flushing should cause an
        # IntegrityError because of two releases of the same name
        pkg.releases.extend(list(releases))

        with py.test.raises(exc.IntegrityError):
            db_session.flush()


def eq_(a, b):
    assert a == b


class TestReleaseUnion(object):

    def test_removes_release_with_duplicate_version(self):
        from ..axle import release_union

        releases = (models.Release(version=u'1.2.4'),
                    models.Release(version=u'1.2.4'))

        releases = release_union(releases)

        assert 1 == len(list(releases))

    def test_retains_releases_with_distinct_versions(self):
        from ..axle import release_union

        releases = (models.Release(version=u'1.2.4'),
                    models.Release(version=u'1.2.5'))
        expected = [u'1.2.4', u'1.2.5']

        releases = release_union(releases)

        for expect, got in itertools.izip_longest(expected, releases):
            yield eq_, expect, got.version

    def test_prefers_releases_with_ids(self, db_session):
        from ..axle import release_union

        pkg = models.Package(name=u'belt')
        pkg.releases.append(models.Release(version=u'1.2.4'))

        releases = (models.Release(version=u'1.2.4'),
                    models.Release(version=u'1.2.4'),
                    pkg.releases[0])
        db_session.add(pkg)
        db_session.flush()

        release, = release_union(releases)
        assert release.id is not None



import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


requires = [
    'pyramid',
    'pyramid_jinja2',
    'SQLAlchemy',
    'transaction',
    'pyramid_tm',
    'zope.sqlalchemy',
    'pyramid_debugtoolbar',
    'waitress',
    'logsna',
    'wheel',
    'distribute',
    'psycopg2',
]

test_requires = [
    'fudge',
    'httpretty',
    'pytest-cov',
    'pytest',
]

setup(name='belt',
      version='0.5',
      description='belt',
      license='BSD',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
          "Programming Language :: Python",
          "Framework :: Pyramid",
          "License :: OSI Approved :: BSD License",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
      ],
      author='Rob Berry',
      author_email='',
      url='',
      keywords='web wsgi bfg pylons pyramid',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='belt',
      install_requires=requires,
      tests_require=test_requires,
      cmdclass={'test': PyTest},
      entry_points="""\
      [paste.app_factory]
      main = belt:main
      [console_scripts]
      createwheels = belt.scripts.createwheels:_build_wheels
      initialize_belt_db = belt.scripts.initializedb:main
      """,
      )



import re
import os
import errno
import shutil
import logging
import urllib2
import xmlrpclib
import subprocess
import collections
from wheel.install import WHEEL_INFO_RE


logger = logging.getLogger(__name__)


ARCHIVE_SUFFIX = re.compile(
    r"(\.zip|\.tar\.gz|\.tgz|\.tar\.bz2|-py[23]\.\d-.*|\.win-amd64-py[23]\.\d\..*|\.win32-py[23]\.\d\..*)$",
    re.IGNORECASE)


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as err:
        if err.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def split_package_name(name):

    wheel_match = WHEEL_INFO_RE(name)
    if wheel_match:
        pkg_name = wheel_match.group('name')
        version = wheel_match.group('ver')
    else:
        pkg_name = re.split(r'-\d+', name, 1)[0]
        version = name[len(pkg_name) + 1:]
        version = ARCHIVE_SUFFIX.sub('', version)
    return pkg_name, version


def copy_wheels_to_pypi(wheel_dir, local_pypi):

    for wheel in os.listdir(wheel_dir):
        match = WHEEL_INFO_RE(wheel)
        if match is None:
            continue
        tags = match.groupdict()
        path_to_wheel = os.path.abspath(os.path.join(wheel_dir, wheel))

        package_dir = os.path.join(local_pypi, tags['name'])
        local_wheel = os.path.join(package_dir, os.path.basename(path_to_wheel))

        # TODO Should we create md5s as we do with source packages?
        if os.path.exists(local_wheel):
            continue

        mkdir_p(package_dir)
        shutil.copyfile(path_to_wheel, local_wheel)


def build_wheels(local_pypi, wheel_dir):

    built_wheels = []
    for wheel in os.listdir(wheel_dir):
        match = WHEEL_INFO_RE(wheel)
        if match is None:
            continue
        tags = match.groupdict()
        # path_to_wheel = os.path.abspath(os.path.join(wheel_dir, wheel))
        built_wheels.append(tags['namever'])

    for package_dir in os.listdir(local_pypi):
        dir_ = os.path.join(local_pypi, package_dir)
        for pkg in os.listdir(dir_):
            name = os.path.join(dir_, pkg)
            _, ext = os.path.splitext(name)
            if ext in ['.md5', '.whl']:
                continue
            package_name, version = split_package_name(pkg)
            package_name = '{}-{}'.format(package_name, version)
            if package_name in built_wheels:
                continue
            args = 'pip wheel --wheel-dir {} {}'.format(wheel_dir, name)
            subprocess.call(args, shell=True)


def get_xmlrpc_client():
    return xmlrpclib.ServerProxy('https://pypi.python.org/pypi')


def get_package_name(name, client=None):
    client = client or get_xmlrpc_client()
    r = urllib2.Request('http://pypi.python.org/simple/{0}'.format(name))
    return urllib2.urlopen(r).geturl().split('/')[-2]


def package_releases(package, location=None, client=None):
    from belt import models
    client = client or get_xmlrpc_client()
    logger.debug('Obtaining releases for ' + package)
    for version in client.package_releases(package, True):
        logger.debug('Found {}-{}'.format(package, version))
        for pkg_data in client.release_urls(package, version):
            rel = models.Release(version=version)
            rel_file = models.File(md5=pkg_data['md5_digest'],
                                   download_url=pkg_data['url'],
                                   location=location or u'',
                                   filename=pkg_data['filename'],
                                   kind=pkg_data['python_version'])
            rel.files.append(rel_file)
            yield rel


def release_union(releases):
    keep = collections.defaultdict(list)
    for rel in releases:
        keep[rel.version].append(rel)

    for v in keep.values():
        yield sorted(v, key=lambda i: i.id, reverse=True)[0]


def create_release(version, url):
    from belt import models
    from belt.utils import get_package_from_pypi
    pkg = get_package_from_pypi(url)
    rel = models.Release(version=version, download_url=url)



from belt.axle import build_wheels, copy_wheels_to_pypi


def _build_wheels():
    import argparse
    parser = argparse.ArgumentParser(description='Build wheels from packages')
    parser.add_argument('--wheel-dir', help='Directory to store wheels in',
                        required=True)
    parser.add_argument('--pypi-dir',
                        help='Directory packages can be found in',
                        required=True)
    args = parser.parse_args()
    build_wheels(args.pypi_dir, args.wheel_dir)
    copy_wheels_to_pypi(args.wheel_dir, args.pypi_dir)


if __name__ == "__main__":
    _build_wheels()



import os
import urllib2
import logging
from pyramid.view import view_config, notfound_view_config
from pyramid.i18n import TranslationStringFactory
from pyramid.response import FileResponse
from pyramid.httpexceptions import (HTTPNotFound,
                                    HTTPMovedPermanently, status_map)

from .utils import (get_package, pypi_url, get_package_from_pypi,
                    store_locally, convert_url_to_pypi)

from sqlalchemy.orm import exc
from .models import DBSession, Package, File, Release, get_or_create
from .axle import (split_package_name, package_releases)

_ = TranslationStringFactory('belt')


PYPI_BASE_URL = 'https://pypi.python.org/packages'


log = logging.getLogger(__name__)


def download_exception(exc):
    if hasattr(exc, 'reason'):
        code = 503
    if hasattr(exc, 'code'):
        code = exc.code
    return status_map[code]


def download_error_msg(exc, requested_package, requested_url):
    if hasattr(exc, 'reason'):
        msg = (u'Error connecting to pypi and retrieving {} ({}) because "{}"'
               .format(requested_package, requested_url, exc.reason))
    elif hasattr(exc, 'code'):
        msg = (u'Pypi could not return {}. Error code: {}'
               .format(requested_package, exc.code))
    else:
        msg = u'Failed to download ' + requested_package
    return msg


@notfound_view_config(append_slash=True)
def notfound(request):
    return HTTPNotFound()


@view_config(route_name='simple', renderer='simple.html')
def simple_list(request):
    pkgs = DBSession.query(Package).all()
    return {'packages': pkgs}


@view_config(route_name='package_list', renderer='package_list.html')
def package_list(request):
    name = request.matchdict['package']
    package_dir = request.registry.settings['local_packages']
    try:
        pkg = Package.by_name(name)
    except exc.NoResultFound:
        pkg = Package.create_from_pypi(name=name, package_dir=package_dir)

    if not pkg.releases:
        releases = {}
        for rel in package_releases(pkg.name,
                                    location=os.path.join(package_dir, pkg.name)):
            release = releases.setdefault(rel.version, rel)
            pkg.releases.append(release)
        DBSession.add(pkg)

    if name != pkg.name:
        dest = request.route_url('package_list', package=pkg.name)
        log.info('Redirecting from /{} to /{}'.format(name, pkg.name))
        return HTTPMovedPermanently(dest)

    return {'package': pkg,
            'kind': 'source',
            'letter': name[0],
            'package_name': name}


@view_config(route_name='package_version', renderer='package_list.html')
def package_version(request):
    name = request.matchdict['package']
    version = request.matchdict['version']

    # when specifiying a version pip requests /simple/package/version but it
    # seems that url scheme never resolves on the real pypi which gives us a
    # chance to inject a decision (perhaps that is the purpose of that
    # lookup). If we have a local package with said name and version we can
    # redirect to our local list, else use the pypi list for this package in
    # case it exists there. Unfortunately we cannot just directly download
    # from this point as we don't have the 'kind' value
    package_dir = request.registry.settings['local_packages']
    package_path = get_package(package_dir, name, version)
    if not package_path.exists:
        log.info(package_path.path + u' does not exist, retrieve from pypi')
    else:
        log.info(package_path.path + u' exists')
    return HTTPNotFound()


@view_config(route_name='download_package')
def download_package(request):
    package, basename, kind, letter = [request.matchdict[key] for key in
                                       ['package', 'basename', 'kind', 'letter']]
    name, version = split_package_name(basename)
    ask_pypi = False

    try:
        rel_file = File.by_filename(basename)
    except exc.NoResultFound:
        package_dir = request.registry.settings['local_packages']
        package_dir = os.path.join(package_dir, package)
        rel_file = File(location=package_dir, filename=basename,
                        download_url=convert_url_to_pypi(request.path))

        try:
            rel = Release.for_package(name, version)
        except exc.NoResultFound:
            rel = Release(version=version)
        rel.files.append(rel_file)
        if not rel.package:
            pkg, _ = get_or_create(DBSession, Package, name=name)
            rel.package = pkg
        DBSession.add(rel)
        ask_pypi = True

    if ask_pypi or not os.path.exists(rel_file.fullpath):

        try:
            fo = get_package_from_pypi(rel_file.download_url)
        except urllib2.URLError as err:
            log.exception(download_error_msg(err, package,
                                             rel_file.download_url))
            raise download_exception(err)
        store_locally(rel_file.fullpath, fo)
    return FileResponse(rel_file.fullpath)



import os
import pytest
from belt import models


class TestSeedPackages(object):

    def test_handles_zip_files(self, tmpdir):
        from ..models import seed_packages
        quux = tmpdir.mkdir('quux').join('quux-2.0.zip')
        quux.write('')

        pkg, = seed_packages(str(tmpdir))
        expected = os.path.join(str(tmpdir), 'quux', 'quux-2.0.zip')
        rel, = pkg.releases
        rel_file, = rel.files
        assert expected == rel_file.fullpath

    def test_sets_filename_for_whl_files(self, tmpdir):
        from ..models import seed_packages
        quux = tmpdir.mkdir('quux').join('quux-1.1-py27-none-any.whl')
        quux.write('')

        pkg, = seed_packages(str(tmpdir))
        expected = os.path.join(str(tmpdir), 'quux',
                                'quux-1.1-py27-none-any.whl')
        rel, = pkg.releases
        rel_file, = rel.files
        assert expected == rel_file.fullpath

    def test_detects_version_for_whl_files(self, tmpdir):
        from ..models import seed_packages
        quux = tmpdir.mkdir('quux').join('quux-1.1-py27-none-any.whl')
        quux.write('')

        pkg, = seed_packages(str(tmpdir))
        rel, = pkg.releases
        assert u'1.1' == rel.version

    def test_groups_releases(self, tmpdir):
        from ..models import seed_packages
        quux = tmpdir.mkdir('quux').join('quux-1.1-py27-none-any.whl')
        quux.write('')

        quux = tmpdir.join('quux', 'quux-1.1.zip')
        quux.write('')

        pkg, = seed_packages(str(tmpdir))
        assert 1 == len(pkg.releases)

    def test_release_can_have_many_files(self, tmpdir):
        from ..models import seed_packages
        quux = tmpdir.mkdir('quux').join('quux-1.1-py27-none-any.whl')
        quux.write('')

        quux = tmpdir.join('quux', 'quux-1.1.zip')
        quux.write('')

        pkg, = seed_packages(str(tmpdir))
        release, = pkg.releases
        assert 2 == len(release.files)


@pytest.fixture
def session(request):
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://', echo=False)

    from sqlalchemy.orm import scoped_session, sessionmaker
    Session = scoped_session(sessionmaker())

    models.Base.metadata.create_all(engine)
    Session.configure(bind=engine)

    connection = engine.connect()
    trans = connection.begin()

    def fin():
        trans.rollback()
    request.addfinalizer(fin)
    return Session
